#ifndef FrameCPP_VERSION_4_FrStatData_HH
#define FrameCPP_VERSION_4_FrStatData_HH

#include <vector>

#include "framecpp/Common/SearchContainer.hh"
#include "framecpp/Common/TOCInfo.hh"
#include "framecpp/Common/FrStatData.hh"

#include "framecpp/Version3/FrStatData.hh"

#include "framecpp/Version4/FrameSpec.hh"
#include "framecpp/Version4/STRING.hh"
#include "framecpp/Version4/PTR_STRUCT.hh"

#include "framecpp/Version4/FrDetector.hh"
#include "framecpp/Version4/FrSE.hh"
#include "framecpp/Version4/FrTable.hh"
#include "framecpp/Version4/FrTOC.hh"
#include "framecpp/Version4/FrVect.hh"

namespace FrameCPP
{
  namespace Version_4
  {
    //===================================================================
    //===================================================================
    class FrStatDataStorage
    {
    public:
      typedef INT_4U timeStart_type;
      typedef INT_4U timeEnd_type;
      typedef INT_4U version_type;

      FrStatDataStorage( );

      FrStatDataStorage( const std::string& Name,
			 const std::string& Comment,
			 const std::string& Representation,
			 const timeStart_type TimeStart,
			 const timeEnd_type TimeEnd,
			 const version_type Version );

      const std::string& GetName( ) const;
      const std::string& GetComment( ) const;
      const std::string& GetRepresentation( ) const;
      timeStart_type GetTimeStart( ) const;
      timeEnd_type GetTimeEnd( ) const;
      version_type GetVersion( ) const;

    protected:
      struct data_type
      {
	STRING		name;
	STRING		comment;
	STRING		representation;
	timeStart_type	timeStart;
	timeEnd_type	timeEnd;
	version_type	version;

	cmn_streamsize_type Bytes( ) const;
	static void Describe( Common::Description& Desc );

	bool operator==( const data_type& RHS ) const;
	void operator( )( Common::IStream& Stream );
	void operator( )( Common::OStream& Stream ) const;
	void operator( )( const FrStatDataStorage& Source );
	void operator( )( const Previous::FrStatDataStorage& Source );

      };

      data_type	m_data;
    };

    inline cmn_streamsize_type FrStatDataStorage::data_type::
    Bytes( ) const
    {
      return
	name.Bytes( )
	+ comment.Bytes( )
	+ representation.Bytes( )
	+ sizeof( timeStart )
	+ sizeof( timeEnd )
	+ sizeof( version )
	;
    }

    inline void FrStatDataStorage::data_type::
    Describe( Common::Description& Desc )
    {
      Desc( FrSE( "name", "STRING" ) );
      Desc( FrSE( "comment", "STRING" ) );
      Desc( FrSE( "representation", "STRING" ) );
      Desc( FrSE( "timeStart", "INT_4U" ) );
      Desc( FrSE( "timeEnd", "INT_4U" ) );
      Desc( FrSE( "version", "INT_4U" ) );
    }

    inline bool FrStatDataStorage::data_type::
    operator==( const data_type& RHS ) const
    {
      return ( ( this == &RHS )
	       || ( ( name == RHS.name )
		    && ( comment == RHS.comment )
		    && ( representation == RHS.representation )
		    && ( timeStart == RHS.timeStart )
		    && ( timeEnd == RHS.timeEnd )
		    && ( version == RHS.version )
		    ) )
	;
    }

    inline void FrStatDataStorage::data_type::
    operator( )( Common::IStream& Stream )
    {
      Stream >> name
	     >> comment
	     >> representation
	     >> timeStart
	     >> timeEnd
	     >> version
	;
    }

    inline void FrStatDataStorage::data_type::
    operator( )( Common::OStream& Stream ) const
    {
      Stream << name
	     << comment
	     << representation
	     << timeStart
	     << timeEnd
	     << version
	;
    }

    inline void FrStatDataStorage::data_type::
    operator( )( const FrStatDataStorage& Source )
    {
      name = Source.GetName( );
      comment = Source.GetComment( );
      representation = Source.GetRepresentation( );
      timeStart = Source.GetTimeStart( );
      timeEnd = Source.GetTimeEnd( );
      version = Source.GetVersion( );
    }

    inline void FrStatDataStorage::data_type::
    operator( )( const Previous::FrStatDataStorage& Source )
    {
      name = Source.GetName( );
      comment = Source.GetComment( );
      timeStart = Source.GetTimeStart( );
      timeEnd = Source.GetTimeEnd( );
      version = Source.GetVersion( );
    }

    inline FrStatDataStorage::
    FrStatDataStorage( )
    {
    }

    inline FrStatDataStorage::
    FrStatDataStorage( const std::string& Name,
		       const std::string& Comment,
		       const std::string& Representation,
		       const timeStart_type TimeStart,
		       const timeEnd_type TimeEnd,
		       const version_type Version )
    {
      m_data.name = Name;
      m_data.comment = Comment;
      m_data.representation = Representation;
      m_data.timeStart = TimeStart;
      m_data.timeEnd = TimeEnd;
      m_data.version = Version;
    }

    inline const std::string& FrStatDataStorage::
    GetName( ) const
    {
      return m_data.name;
    }

    inline const std::string& FrStatDataStorage::
    GetComment( ) const
    {
      return m_data.comment;
    }

    inline const std::string& FrStatDataStorage::
    GetRepresentation( ) const
    {
      return m_data.representation;
    }

    inline FrStatDataStorage::timeStart_type FrStatDataStorage::
    GetTimeStart( ) const
    {
      return m_data.timeStart;
    }

    inline FrStatDataStorage::timeEnd_type FrStatDataStorage::
    GetTimeEnd( ) const
    {
      return m_data.timeEnd;
    }

    inline FrStatDataStorage::version_type FrStatDataStorage::
    GetVersion( ) const
    {
      return m_data.version;
    }

    //===================================================================
    //===================================================================
    template< typename TDetector, typename TVect, typename TTable >
    class FrStatDataRefsT
    {
    public:
      static const INT_2U MAX_REF = 3;

      typedef TDetector detector_type;

      typedef typename Common::SearchContainer< TVect, &TVect::GetName >
      data_type;
      typedef typename data_type::iterator iterator;
      typedef typename data_type::const_iterator const_iterator;

      typedef typename Common::SearchContainer< TTable, &TTable::GetNameSlow >
      table_type;
      typedef typename table_type::iterator table_iterator;
      typedef typename table_type::const_iterator const_table_iterator;

      TDetector* GetDetector( ) { return m_refs.detector; }
      const TDetector* GetDetector( ) const { return m_refs.detector; }
      const data_type&	RefData( ) const { return m_refs.data; }
      data_type& RefData() { return m_refs.data; }
      const table_type&	RefTable() const { return m_refs.table; }
      table_type& RefTable() { return m_refs.table; }

      inline detector_type**
      AddressOfDetector( ) const
      {
	return const_cast< detector_type**>( &m_refs.detector );
      }

      inline void
      SetDetector( TDetector* Detector )
      {
	m_refs.detector = Detector;
      }   

    protected:
      struct ref_type
      {
	detector_type*	detector;
	data_type	data;
	table_type	table;

	inline cmn_streamsize_type
	Bytes( const Common::StreamBase& Stream ) const
	{
	  return
	    Stream.PtrStructBytes( )	// detector
	    + Stream.PtrStructBytes( )	// data
	    + Stream.PtrStructBytes( )	// table
	    ;
	}

	static inline void
	Describe( Common::Description& Desc )
	{
	  Desc( FrSE( "detector", PTR_STRUCT::Desc( detector_type::StructName( ) ) ) );
	  Desc( FrSE( "data", PTR_STRUCT::Desc( TVect::StructName( ) ) ) );
	  Desc( FrSE( "table", PTR_STRUCT::Desc( TTable::StructName( ) ) ) );
	}

	inline bool
	operator==( const ref_type& RHS ) const
	{
	  return ( ( this == &RHS )
		   || ( ( ( detector == RHS.detector )
			  || ( detector
			       && RHS.detector
			       && ( *detector == *(RHS.detector) )
			       ) )
			&& ( data == RHS.data )
			&& ( table == RHS.table )
			) )
	    ;
	}

	inline void
	operator( )( Common::IStream& Stream )
	{
	  void* d = &detector;

	  Stream >> static_cast< Common::FrameSpec::Object** >( d )
		 >> data
		 >> table
	    ;
	}

	inline void
	operator( )( Common::OStream& Stream ) const
	{
	  Stream << static_cast< const Common::FrameSpec::Object* >( detector )
		 << data
		 << table
	    ;
	}
      };

      ref_type	m_refs;
    };

    typedef FrStatDataRefsT< FrDetector, FrVect, FrTable > FrStatDataRefs;

    //===================================================================
    //===================================================================
    class FrStatData
      : public Common::FrStatData,
	public FrStatDataStorage,
	public FrStatDataRefs,
	public Common::TOCInfo
    {
    public:
      static const INT_4U ALL_VERSIONS;
      static const INT_4U LATEST_VERSION;

#if WORKING
      class Query
      {
      public:
	~Query( );
	void Add( const std::string& NamePattern,
		  const std::string& DetectorPattern,
		  const General::GPSTime& StartTime,
		  const General::GPSTime& EndTime,
		  const INT_4U Version );

	void Reset( Common::IStream& Stream );
	INT_4U size( ) const;

	FrStatData* operator[]( INT_4U Index );
      private:

	struct query_info_type {
	  FrStatData*	s_fr_stat_data;
	  INT_8U	s_fr_stat_pos;
	  bool		s_own;
	  // Information needed when looking for the latest version
	  std::string	s_nameStat;
	  std::string	s_detector;
	  INT_4U	s_tStart;
	  INT_4U	s_tEnd;
	  INT_4U	s_version;
	};

	typedef std::vector< query_info_type >	data_type;

	class CompareFilePosition {
	public:
	  int operator( )( const query_info_type& LHS,
			   const query_info_type& RHS ) const;
	};

	std::list< const FrDetector* >	m_detectors;
	bool				m_dirty;
	data_type			m_fr_stat_data;
	Common::IStream* 		m_stream;

	void add( const FrTOC::StatTypeKey& Key, 
		  const FrTOC::StatType_type& Data,
		  INT_4U Start,
		  INT_4U Stop,
		  INT_4U Version );

	void load( );
	std::auto_ptr< FrStatData >
	read_fr_stat_data( Common::IStream& Stream,
			   const std::string& Detector,
			   INT_8U PositionOfStatData );

      };
#endif /* WORKING */

      //: Default constructor
      FrStatData( );
      FrStatData( const FrStatData& statData );
      FrStatData( const std::string& name, const std::string& comment,
		  const std::string& representation,
		  INT_4U timeStart,INT_4U timeEnd,
		  INT_4U version );
      FrStatData( const Previous::FrStatData& Source,
		  Common::IStream* Stream );

      static std::auto_ptr< Object >
      Promote( INT_2U Source,
	       Object* Obj, Common::IStream* Stream );

      static const char* const StructName( );

      static const Common::Description* StructDescription( );

      //: Number of bytes needed to write this structure
      cmn_streamsize_type Bytes( const Common::StreamBase& Stream) const;

      virtual FrStatData* Clone( ) const;

      virtual FrStatData* Create( Common::IStream& Stream ) const;

      virtual const FrDetector* GetDetector( ) const
      {
	return FrStatDataRefs::GetDetector( );
      }

      virtual const char* const ObjectStructName( ) const;

      virtual void Write( Common::OStream& Stream ) const;

      virtual void TOCQuery( int InfoClass, ... ) const;

      //: Merge with another FrAdcData
      FrStatData& Merge( const FrStatData& RHS );

      /// \brief comparison operator
      bool operator==( const FrStatData& RHS ) const;

      /// \brief comparison operator
      virtual bool operator==( const Common::FrameSpec::Object& Obj ) const;

    protected:
      /// \brief Down grade an object
      virtual std::auto_ptr< Object >
      demote( INT_2U Target, Common::IStream* Stream ) const;

      /// \brief Upgrade an object
      virtual std::auto_ptr< Object >
      promote( INT_2U Target, Object* Obj, Common::IStream* Stream ) const;

    private:
      using Common::FrameSpec::Object::Create;

      FrStatData( Common::IStream& Stream );
    };

    //===================================================================
    //===================================================================
#if WORKING
    inline int FrStatData::Query::CompareFilePosition::
    operator( )( const query_info_type& LHS,
		 const query_info_type& RHS ) const
    {
      return LHS.s_fr_stat_pos < RHS.s_fr_stat_pos;
    }

    inline INT_4U FrStatData::Query::
    size( ) const
    {
      return m_fr_stat_data.size( );
    }
#endif /* WORKING */


    //===================================================================
    //===================================================================
    inline std::auto_ptr< Common::FrameSpec::Object > FrStatData::
    Promote( INT_2U Source,
	     Common::FrameSpec::Object* Obj, Common::IStream* Stream )
    {
      return Object::PromoteObject< Previous::FrStatData, FrStatData >
	( DATA_FORMAT_VERSION, Source, Obj, Stream );
    }

    inline const char* const FrStatData::
    StructName( )
    {
      static const char* const name = "FrStatData";
      return name;
    }
  } // namespace - Version_4
} // namespace FrameCPP
#endif /* FrameCPP_VERSION_4_FrStatData_HH */
