#ifndef FrameCPP_VERSION_6_FrTOCAdcData_HH
#define FrameCPP_VERSION_6_FrTOCAdcData_HH

#include "framecpp/Version6/STRING.hh"

namespace FrameCPP
{
  namespace Common
  {
    class TOCInfo;
  }

  namespace Version_6
  {
    //===================================================================
    //===================================================================
    class FrTOCAdcData
    {
    public:
      typedef INT_4U	nadc_type;
      typedef STRING	name_type;
      typedef INT_4U	channel_id_type;
      typedef INT_4U	group_id_type;
      typedef INT_8U	position_type;

      struct adc_info_type {
	channel_id_type			m_channelID;
	group_id_type			m_groupID;
	std::vector< position_type >	m_positionADC;
      };

      typedef std::map< name_type, adc_info_type >	MapADC_type;

      FrTOCAdcData( );

      Common::FrameSpec::size_type
      Bytes( const Common::StreamBase& Stream ) const;

      const MapADC_type& GetADC( ) const;

      MapADC_type::const_iterator GetADC( const std::string& Name ) const;

      MapADC_type::const_iterator GetADC( INT_4U index ) const;

      void QueryAdc( const Common::TOCInfo& Info,
		     INT_4U FrameOffset,
		     INT_8U Position );

      bool operator==( const FrTOCAdcData& RHS ) const;

      static void Description( Common::Description& Desc );

    protected:
      typedef std::vector< name_type >	key_container_type;
      key_container_type	m_keys;
      MapADC_type		m_info;

      FrTOCAdcData( Common::IStream& Stream, INT_4U FrameCount );
      void write( Common::OStream& Stream ) const;
    };

    inline Common::FrameSpec::size_type FrTOCAdcData::
    Bytes( const Common::StreamBase& Stream ) const
    {
      Common::FrameSpec::size_type 
	retval = sizeof( nadc_type );
      if ( m_info.size( ) )
      {
	retval
	  += ( m_info.size( )
	       * ( sizeof( channel_id_type )
		   + sizeof( group_id_type )
		   + ( sizeof( position_type )
		       * m_info.begin( )->second.m_positionADC.size( ) )
		   ) );
      }
      for ( MapADC_type::const_iterator
	      cur = m_info.begin( ),
	      last = m_info.end( );
	    cur != last;
	    ++cur )
      {
	retval += cur->first.Bytes( );
      }
      return retval;
    }

    inline const FrTOCAdcData::MapADC_type& FrTOCAdcData::
    GetADC( ) const
    {
      return m_info;
    }

  } // namespace - Version_6
} // namespace - FrameCPP

#endif /* FrameCPP_VERSION_6_FrTOCAdcData_HH */
